#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "event.h"
#include "pxy.h"

int pxytimeout = 20000; /* 20 sec */
unsigned char pxybuf[8193];

static void timedout(int UNUSED fd, ev_t UNUSED e, void *ctx) {
  errno = ETIMEDOUT;
  ev_iowakeup((int)ctx, EVENT_ERROR);
}

int pxyrenew(int fd, pxyconn_t *c, int tmo) {
  if (ev_settimer(tmo ? tmo : pxytimeout, timedout, (void*)fd) == 0)
    return 0;
  else {
    pxyaction(fd, c, -1);
    return -1;
  }
}

int pxycerr(int fd, ev_t e, pxyconn_t *c) {
  if (e & EVENT_RDWR)
    return 0;
  /*
  if (errno != ETIMEDOUT)
    pxyinfo(c, 1, "error: %s", strerror(errno));
  */
  pxyaction(fd, c, -1);
  return -1;
}

pxyconn_t *pxynew() {
  pxyconn_t *c = (pxyconn_t *)malloc(sizeof(pxyconn_t));
  if (c)
    memset(c, 0, sizeof(*c));
  return c;
}

void pxyfree(int fd, pxyconn_t *c) {
  if (fd >= 0) {
    ev_settimer(-1, timedout, (void*)fd);
    ev_reqio(fd, 0, NULL, NULL);
  }
  if (c->buf) {
    free(c->buf);
    c->buf = NULL;
  }
  free(c);
}

static int
pxyev_reqio(int fd, ev_t e, void(*fn)(int,ev_t,pxyconn_t*), pxyconn_t *c) {
  if (ev_reqio(fd, e, (void(*)(int,ev_t,void*))fn, c) == 0)
    return 0;
  else {
    pxyaction(fd, c, -1);
    return -1;
  }
}

int pxyvwrite(int fd, pxyconn_t *c, const char *buf, int len) {
  if (write(fd, buf, len) < 0) {
    pxyaction(fd, c, -1);
    return -1;
  }
  else {
    pxyvio(c, 1, buf, buf + len);
    return 0;
  }
}

int pxystart(int fd, pxyconn_t *c) {
  return
    pxyev_reqio(fd, EVENT_WRITE, c->proto->handler, c) == 0 &&
    pxyrenew(fd, c, 0) == 0 ?
    0 : -1;
}

int pxyread(int fd, pxyconn_t *c, char *buf, int l) {
  errno = 0;
  l = read(fd, buf, l);
  if (l <= 0) {
    pxyaction(fd, c, l);
    return -1;
  }
  if (l > 0) {
    c->nread += l;
    buf[l] = '\0';
  }
  return l;
}

int pxyreadnext(int fd, pxyconn_t *c) {
  int l, r;
  if (c->buf && (l = c->buf->len) != 0)
    memcpy(pxybuf, c->buf->buf, l);
  else
    l = 0;
  r = pxyread(fd, c, pxybuf + l, sizeof(pxybuf) - 1 - l);
  return r <= 0 ? r : r + l;
}

int pxysave(int fd, pxyconn_t *c, char *buf, unsigned len, unsigned max) {
  if (c->nread > 2048) {
    pxyinfo(c, 2, "too much input");
    pxyaction(fd, c, 0);
    return -1;
  }
  if (buf && len > 0) {
    if (!c->buf && !(c->buf = (pxybuf_t*)malloc(sizeof(pxybuf_t)))) {
      errno = ENOMEM;
      pxyaction(fd, c, -1);
      return -1;
    }
    if (len > sizeof(c->buf->buf)) {
      buf += len - sizeof(c->buf->buf);
      len = sizeof(c->buf->buf);
    }
    if (max && len > max) {
      buf += len - max;
      len = max;
    }
    memcpy(c->buf->buf, buf, len);
    c->buf->len = len;
  }
  else if (c->buf)
    c->buf->len = 0;
  return 0;
}

static void
hch(int fd, ev_t e, pxyconn_t *c) {
  /* HTTP CONNECT.
   * state:
   *  1 - request sent, want reply
   */
  int l;
  if (pxycerr(fd, e, c)) return;
  if (!c->state) {
    l = sprintf(pxybuf, "CONNECT %s:%d HTTP/1.0\r\n\r\n",
		inet_ntoa(c->dst.addr), c->dst.port);
    if (pxyvwrite(fd, c, pxybuf, l) == 0) {
      c->state = 1;
      if (pxyrenew(fd, c, 0) == 0)
	pxyev_reqio(fd, EVENT_READ, hch, c);
    }
  }
  else 
    pxyaction(fd, c, 1);
}

static void
hcc(int fd, pxyconn_t *c, char *buf, int l) {
  if (c->nread == l && memmem(buf, l, "HwaRang/", 8)) {
    pxyinfo(c, 2, "probably HwaRang proxy, workaround enabled");
    write(fd, "\r\n", 2);
  }
}

static void
s5h(int fd, ev_t e, pxyconn_t *c) {
  /* SOCKS5
   * state:
   *  1 - auth request sent, want reply
   *  2 - connect request sent, want reply
   */
  int l;
  if (pxycerr(fd, e, c)) return;
  if (!c->state) {
    pxybuf[0] = 5; /* socks version */
    pxybuf[1] = 1; /* nmethods */
    pxybuf[2] = 0; /* no auth reqd */
    if (write(fd, pxybuf, 3) < 0)
      pxyaction(fd, c, -1);
    else {
      c->state = 1;
      if (pxyev_reqio(fd, EVENT_READ, s5h, c) == 0)
	pxyrenew(fd, c, pxytimeout / 2);
    }
  }
  else if (c->state == 1) {
    if ((l = pxyread(fd, c, pxybuf, 4)) < 2) {
      if (l > 0)
	pxyaction(fd, c, -1);
    }
    else if (pxybuf[1] != 0) {
      pxyinfo(c, 2, "auth rejected (v=%d c=%d%s)",
	      pxybuf[0], pxybuf[1],
	      pxybuf[0] == 0 && pxybuf[1] == 91 ?
	        ", socks4-only server" : "");
      pxyaction(fd, c, 0);
    }
    else if (l > 2) {
      pxyinfo(c, 2, "too much socks5 data v=%d c=%d %d %d",
	      pxybuf[0], pxybuf[1], pxybuf[2], pxybuf[3]);
      pxyaction(fd, c, 0);
    }
    else {
      ipport_t p = htons(c->dst.port);
	pxyinfo(c, 3, "auth accepted");
      pxybuf[0] = 5; /* version */
      pxybuf[1] = 1; /* connect cmd */
      pxybuf[2] = 0; /* reserved */
      pxybuf[3] = 1; /* ipv4 addrtype */
      memcpy(pxybuf+4, &c->dst.addr.s_addr, 4);
      memcpy(pxybuf+8, &p, 2);
      if (write(fd, pxybuf, 10) < 0)
	pxyaction(fd, c, -1);
      else {
	c->state = 2;
	pxyrenew(fd, c, 0);
      }
    }
  }
  else if ((l = pxyread(fd, c, pxybuf, 10)) < 2) {
    if (l > 0)
      pxyaction(fd, c, -1);
  }
  else if (pxybuf[1] == 0) {
    pxyinfo(c, 2, "request granted (v=%d c=%d l=%d)",
	    pxybuf[0], pxybuf[1], l);
    pxyaction(fd, c, 1);
  }
  else {
    char *s;
    switch(pxybuf[1]) {
    case 1: s = "general server failure"; break;
    case 2: s = "connection not allowed by ruleset"; break;
    case 3: s = "network unreachable"; break;
    case 4: s = "host unreachable"; break;
    case 5: s = "connection refused"; break;
    case 6: s = "TTL expired"; break;
    case 7: s = "command not supported"; break;
    case 8: s = "address type not supported"; break;
    default: s = "unexpected response"; break;
    }
    pxyinfo(c, 2, "%s (v=%d c=%d)", s, pxybuf[0], pxybuf[1]);
    pxyaction(fd, c, 0);
  }
}

static void
s4h(int fd, ev_t e, pxyconn_t *c) {
  /* SOCKS4.
   * state:
   *  1 - connect request sent, want reply
   */
  int l;
  if (pxycerr(fd, e, c)) return;
  if (!c->state) {
    ipport_t p = htons(c->dst.port);
    pxybuf[0] = 4; /* socks version */
    pxybuf[1] = 1; /* request: connect */
    memcpy(pxybuf+2, &p, 2);
    memcpy(pxybuf+4, &c->dst.addr.s_addr, 4);
    pxybuf[8] = 0; /* empty/null username */
    if (write(fd, pxybuf, 9) < 0)
      pxyaction(fd, c, -1);
    else {
      c->state = 1;
      if (pxyrenew(fd, c, 0) == 0)
	pxyev_reqio(fd, EVENT_READ, s4h, c);
    }
  }
  else if ((l = pxyread(fd, c, pxybuf, 8)) < 8) {
    if (l > 0)
      pxyaction(fd, c, -1);
  }
  else if (pxybuf[1] == 90) {
    pxyinfo(c, 2, "request granted (v=%d c=%d)", pxybuf[0], pxybuf[1]);
    pxyaction(fd, c, 1);
  }
  else {
    char *s;
    switch(pxybuf[1]) {
    case 91: s = "request rejected or failed"; break;
    case 92: s = "identd required"; break;
    case 93: s = "identd info mismatch"; break;
    default: s = "unexpected response"; break;
    }
    pxyinfo(c, 2, "%s (v=%d c=%d)", s, pxybuf[0], pxybuf[1]);
    pxyaction(fd, c, 0);
  }
}

static void
wgh(int fd, ev_t e, pxyconn_t *c) {
  /* WINGATE/telnet.
   * state:
   *  1 - wait for a prompt, interpret Username:/Login: and similar;
   *   we're waiting for a small timeout here, if not found - go to
   *   next state
   *  2 - connection request sent, wait for reply
   */
  int l;
  if (!c->state) {
    if (pxycerr(fd, e, c)) return;
    l = read(fd, pxybuf, 1); /* check if established */
    if (!l) {
      errno = 0;
      pxyaction(fd, c, -1);
    }
    else if (l < 0 && errno != EAGAIN)
      pxyaction(fd, c, -1);
    else {
      c->state = 1;
      if (pxyrenew(fd, c, pxytimeout / 10) == 0)
	pxyev_reqio(fd, EVENT_READ, wgh, c);
    }
  }
  else if (c->state == 1) {
    if (e == EVENT_ERROR && errno == ETIMEDOUT) {
      l = sprintf(pxybuf, "%s:%d\r\n", inet_ntoa(c->dst.addr), c->dst.port);
      if (write(fd, pxybuf, l) < 0)
	pxyaction(fd, c, -1);
      else {
	c->state = 2;
	pxyrenew(fd, c, 0);
      }
    }
    else if (pxycerr(fd, e, c))
      return;
    else if ((l = pxyreadnext(fd, c)) <= 0)
      ;
    else if (memmem(pxybuf, l, "sername", 7) ||
	     memmem(pxybuf, l, "ogin:", 5))
      pxyaction(fd, c, 0);
    else
      pxysave(fd, c, pxybuf, l, 0);
  }
  else if (pxycerr(fd, e, c))
    ;
  else
    pxyaction(fd, c, 1);
}

#if 0
static void
wgc(int fd, pxyconn_t *c, char *buf, int l) {
  /* look at Password:/etc strings? */
}
#endif

static const ipport_t
htp[] = {
  80,81,1182,3128,4480,6588,7033,8000,8080,8081,8085,8090,8095,8100,8105,8110,
  0 },
spp[] = { 1080,1180,0 },
wgp[] = { 23,1181,0 };

const pxyproto_t pxyprotos[] = {
 { "socks5",  "s5", spp, s5h, NULL },
 { "http",    "hc", htp, hch, hcc },
 { "socks4",  "s4", spp, s4h, NULL },
 { "wingate", "wg", wgp, wgh, NULL },
 {0,0,0,0,0}
};

const pxyproto_t *pxyfindproto(const char *name) {
  const pxyproto_t *p = pxyprotos;
  while(p->name)
    if (strcmp(p->name, name) == 0 || strcmp(p->aname, name) == 0)
      return p;
    else
      ++p;
  return NULL;
}
